/****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    buildTrainingImageDataset.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
** DESCRIPTION:
**  Constructs a database of positive and negative training images for
**  training an object detector.
**
*****************************************************************************/

#include <cstdlib>
#include <string>
#include <limits>
#include <iomanip>
#include <map>
#include <set>

// OpenCV library
#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

// stair vision library
#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"

using namespace std;

void usage()
{
    cerr << SVL_USAGE_HEADER << endl;
    cerr << "USAGE: ./buildTrainingImageDataset [OPTIONS] <imageSequence> <objectLabels>?" << endl;
    cerr << "OPTIONS:" << endl
	 << "  -parallel <id> <ct>  :: used for distributed jobs on the cluster." << endl
	 << SVL_STANDARD_OPTIONS_USAGE 
	 << endl;
}


// baseExt is deprecated; will just strip off the extension of every file in the image sequence
// and replace it with the default channel extension
// will iterate through all channels (except edge); if don't want that, should
// use the command line argument to set the channels appropriately in svlImageLoader

// Main ----------------------------------------------------------------------

int main(int argc, char *argv[])
{
  captureOpenCVerrors();

  // read command line parameters
  int parallelId = -1;
  int parallelCount = -1;

  const int NUM_REQUIRED_PARAMETERS = 1;

  string newNameForDepr;

  SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
    SVL_CMDLINE_OPTION_BEGIN("-parallel", ptr)
      parallelId = atoi(ptr[0]);
      parallelCount = atoi(ptr[1]);
      if (parallelId < 0 || parallelId >= parallelCount) {
        SVL_LOG(SVL_LOG_FATAL, "illegal id number " << parallelId);
      } else if (parallelCount < 0) {
        SVL_LOG(SVL_LOG_FATAL, "illegal job count " << parallelCount);
      }
    SVL_CMDLINE_OPTION_END(2)

    // [[TODO: deprecate from here]] --------------------------------------

    else if (!strcmp(*SVL_CMDLINE_ARGV, "-ch")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlImageLoader channels")
      string type = string(*(++SVL_CMDLINE_ARGV));
      string ext = string(*(++SVL_CMDLINE_ARGV));
      pair<string, string> ch = make_pair<string, string>(type, ext);
      svlImageLoader::CHANNELS.push_back(ch);
      SVL_CMDLINE_ARGC -= 2;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-dmsize")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlImageLoader depthMapWidth and depthMapHeight")
      svlImageLoader::DEPTH_MAP_WIDTH = atoi(*(++SVL_CMDLINE_ARGV));
      svlImageLoader::DEPTH_MAP_HEIGHT = atoi(*(++SVL_CMDLINE_ARGV));
      SVL_CMDLINE_ARGC -= 2;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-baseExt")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlImageLoader defaultExtension")
      svlImageLoader::BASE_EXT = *(++SVL_CMDLINE_ARGV);
      SVL_CMDLINE_ARGV--;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-object")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder objects")
      svlTrainingDatasetBuilder::OBJECTS.insert(*(++SVL_CMDLINE_ARGV));
      SVL_CMDLINE_ARGC--;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-aspectRatio")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder negAspectRatio")
      svlTrainingDatasetBuilder::NEG_ASPECT_RATIO = atof(*(++SVL_CMDLINE_ARGV));
      SVL_CMDLINE_ARGC--;
    } 
    else if (!strcmp(*SVL_CMDLINE_ARGV, "-skipPositives")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder skipPos")
      svlTrainingDatasetBuilder::SKIP_POS = true;
    }
    else if (!strcmp(*SVL_CMDLINE_ARGV, "-skipNegatives")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder skipNeg")
      svlTrainingDatasetBuilder::SKIP_NEG = true;
    }
    else if (!strcmp(*SVL_CMDLINE_ARGV, "-posBestWindow")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder posBestWindow")
      svlTrainingDatasetBuilder::POS_BEST_WINDOW = true; 
      assert(!svlTrainingDatasetBuilder::ALL_WINDOWS);
    }
    else if (!strcmp(*SVL_CMDLINE_ARGV, "-allWindows")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder allWindows")
      svlTrainingDatasetBuilder::ALL_WINDOWS = true;
      assert(!svlTrainingDatasetBuilder::POS_BEST_WINDOW);
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-fponly")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder falsePositivesOnly")
      svlTrainingDatasetBuilder::FPOS_ONLY = 1;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-resize")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder resizeWidth and resizeHeight")
      svlTrainingDatasetBuilder::RESIZE_WIDTH = atoi(*(++SVL_CMDLINE_ARGV));
      svlTrainingDatasetBuilder::RESIZE_HEIGHT = atoi(*(++SVL_CMDLINE_ARGV));
      SVL_CMDLINE_ARGC -= 2;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-negatives")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder numNegs")
      svlTrainingDatasetBuilder::NUM_NEG = atoi(*(++SVL_CMDLINE_ARGV));
      SVL_CMDLINE_ARGC--;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-threshold")) {
      SVL_OPTION_DEPRECATED("-threshold", "svlVision.svlTrainingDatasetBuilder threshold")
      svlTrainingDatasetBuilder::THRESHOLD = atof(*(++SVL_CMDLINE_ARGV));
      SVL_CMDLINE_ARGC--;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-includeFlipped")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder includeFlipped")
      svlTrainingDatasetBuilder::INCLUDE_FLIPPED = true;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-includeOthers")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder includeOtherObjects")
      svlTrainingDatasetBuilder::INCLUDE_OTHER_OBJS = true;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-negHeight")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder negHeight")
      svlTrainingDatasetBuilder::NEG_HEIGHT = atoi(*(++SVL_CMDLINE_ARGV));
      SVL_CMDLINE_ARGC--;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-negName")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder negSubdirName")
      svlTrainingDatasetBuilder::NEG_SUBDIR_NAME = *(++SVL_CMDLINE_ARGV);
      SVL_CMDLINE_ARGC--;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-posPrefix")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder posPrefix")
      svlTrainingDatasetBuilder::POS_PREFIX = *(++SVL_CMDLINE_ARGV);
      SVL_CMDLINE_ARGC--;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-negPrefix")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder negPrefix")
      svlTrainingDatasetBuilder::NEG_PREFIX = *(++SVL_CMDLINE_ARGV);
      SVL_CMDLINE_ARGC--;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-createDirs")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder createDirs")
      svlTrainingDatasetBuilder::CREATE_DIRS = true;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-detections")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder detectionsFilename")
      svlTrainingDatasetBuilder::DETECTIONS_FILE = *(++SVL_CMDLINE_ARGV);
      SVL_CMDLINE_ARGC--;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-directory")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder baseDir")
      svlTrainingDatasetBuilder::BASE_DIR = *(++SVL_CMDLINE_ARGV);
      SVL_CMDLINE_ARGC--;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-useWindowReferences")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlTrainingDatasetBuilder useWinRefs")
      svlTrainingDatasetBuilder::USE_WIN_REFS = true;
    } 
      // [[TODO: deprecate to here]] ----------------------------------------

  SVL_END_CMDLINE_PROCESSING(usage())
  
  if (SVL_CMDLINE_ARGC != NUM_REQUIRED_PARAMETERS) {
      usage();
      return -1;
    }

  svlCodeProfiler::tic(svlCodeProfiler::getHandle("buildTrainingImageDataset"));

  const char *imageSequenceFile = SVL_CMDLINE_ARGV[0];
  const char *objectLabelsFile = NULL;
  if (SVL_CMDLINE_ARGC > NUM_REQUIRED_PARAMETERS) 
    objectLabelsFile = SVL_CMDLINE_ARGV[1];

  // encapsulates all the different parameters in a cleaner way
  svlTrainingDatasetBuilder builder(imageSequenceFile, objectLabelsFile);
  builder.writeDataset(parallelCount, parallelId);
  
  svlCodeProfiler::toc(svlCodeProfiler::getHandle("buildTrainingImageDataset"));
  svlCodeProfiler::print(cerr);

  return 0;
}
