/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlHistogramOperators.h
** AUTHOR(S):   Olga Russakovsky <olga@cs.stanford.edu>
** DESCRIPTION:
**  Doesn't maintain the storage for the histogram, but provides
** all the insertion operators. svlHistogram contains just one copy
** of it; svl2dHistogram contains two
** TODO: extend to non-uniform bins
** TODO: extend to have non-zero min
**
*****************************************************************************/

#pragma once

#include <string>
#include <vector>
#include <fstream>
#include <cv.h>
#include "svlBase.h"

using namespace std;

// svlSoftBinAssignment ----------------------------------------------------
// Used to return a soft bin histogram assignment
// TODO: extend to more than two bins potentially
struct svlSoftBinAssignment {
  int _bin[2];
  double _binWeight[2];
  
  svlSoftBinAssignment() { 
    _bin[0] = _bin[1] = 0;
    _binWeight[0] = _binWeight[1] = 0.0;
  }
  svlSoftBinAssignment(int a, double aw, int b, double bw) {
    _bin[0] = a;
    _binWeight[0] = aw;
    _bin[1] = b;
    _binWeight[1] = bw;
  }
};

// svlHistogramOperators ----------------------------------------------------
class svlHistogramOperators
{
 private:
  // set by user
  int _numBins;
  double _minVal;
  bool _circular;

  // precomputed in constructor
  double _binSize;
  double _binCenterOffset;

  // corrects the bin as necessary to account for possible
  // overflow
  inline bool fixBin(int &bin) const {
    if (bin >= _numBins || bin < 0) {
      if (_circular) {
	bin %= _numBins;
	if (bin < 0) bin += _numBins;
      }
      else
	return false;
    }
    return true;
  }

 public:
  svlHistogramOperators(int numBins, double minValue, double maxValue) :
    _numBins(numBins), _minVal(minValue), _circular(false),
    _binSize((maxValue - minValue) / _numBins),
    _binCenterOffset(_binSize/2)
      {
	SVL_ASSERT_MSG(_numBins > 0, "Can't create a histogram with "
		       << _numBins << " bins");
	SVL_ASSERT_MSG(maxValue > minValue, 
		       "Can't create a histogram with range ["
		       << minValue << ", " << maxValue << ")");
      }

  void makeCircular() { _circular = true; }
  
  // return -1 on failure
  inline int getBinNum(double binValue) const {
    int bin = (int)floor((binValue - _minVal)/ _binSize);
    if (!fixBin(bin)) return -1;
    return bin;
  }

  // will alter the value of sb regardless; returns true on success
  inline bool getBinNumLinearSmooth(double binValue,
				    svlSoftBinAssignment &sb) const {
    double bin = (binValue - _minVal - _binCenterOffset) / _binSize;

    // the two bins closest to this value
    sb._bin[0] = (int)floor(bin);
    sb._bin[1] = sb._bin[0] + 1;
    
    // compute the relative weighting factors
    sb._binWeight[0] = 1 - (bin - sb._bin[0]);
    sb._binWeight[1] = 1 - (sb._bin[1] - bin);
    
    // fix the bins to be within range if possible
    bool bin1OK = fixBin(sb._bin[0]);
    bool bin2OK = fixBin(sb._bin[1]);
    
    // if both outside the range, can't do anything
    if (!bin1OK && !bin2OK) return false;
    
    // if just one is outside the range, set the
    // other one to have full weight
    if (!bin1OK) {
      sb._bin[0] = 0;
      sb._binWeight[0] = 0.0;
      sb._binWeight[1] = 1.0;
    }
    if (!bin2OK) {
      sb._bin[1] = 0;
      sb._binWeight[0] = 1.0;
      sb._binWeight[1] = 0.0;
    }

    return true;
  }
};
